#ifndef TEST_JS_NATIVE_API_TEST_CONVERSIONS_TEST_NULL_H_
#define TEST_JS_NATIVE_API_TEST_CONVERSIONS_TEST_NULL_H_

#include <js_native_api.h>

void init_test_null(napi_env env, napi_value exports);

#endif  // TEST_JS_NATIVE_API_TEST_CONVERSIONS_TEST_NULL_H_
