#![allow(non_upper_case_globals)]
#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(unused)]

// This matches bindgen::Builder output
include!(concat!(env!("OUT_DIR"), "/wamr_bindings.rs"));
